# -*- coding: utf-8 -*-

import json
import logging
import random
import re
import requests
import subprocess
from markupsafe import Markup

from odoo import _, api, models, fields
from odoo.addons.openerp_enterprise.models.gandi_api import GANDI, GandiError, TLDS, GTLDS
from odoo.addons.openerp_enterprise.models.public_suffix_list import PublicSuffixList
from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.const import (
    SAAS_BASE_URL, SCHEME, SPECIAL_PREFIXES, TEST, TIMEOUT,
    ABUSE_ALIAS, ABUSE_URLSCAN_APIKEY
)
from odoo.exceptions import AccessDenied, ValidationError
from odoo.tools import lazy


_logger = logging.getLogger(__name__)


PRICE_LIMIT = 30

@lazy
def PSL():
    # lazy load, it opens file (> 50ms)
    return PublicSuffixList()

class DomainRequest(models.Model):
    _name = 'openerp.enterprise.domain.request'
    _description = 'Free domain name request'
    _inherit = ['mail.thread']
    _rec_name = 'fqdn'

    fqdn = fields.Char('FQDN', required=True, index=True, help='Fully Qualified Domain Name')
    user_id = fields.Many2one('res.users', 'Owner', required=True, index=True, ondelete='cascade')
    database_id = fields.Many2one('openerp.enterprise.database', required=True, index=True)
    state = fields.Selection([
        ('pending', 'Pending'),
        ('granted', 'Granted'),
        ('denied', 'Denied'),
    ], string='State', required=True, default='pending', tracking=True)
    payload = fields.Text(required=True, help='JSON payload')

    @api.model
    def _submit(self, fqdn, database, owner_data):
        """
        Submits a new free domain name request for the support to review.
        """
        submitted_requests = self.sudo().search([
            ('database_id', '=', database.id),
            ('state', 'in', ['pending', 'granted']),
        ])
        if submitted_requests:
            if any(r.state == 'pending' for r in submitted_requests):
                raise ValidationError(_('You already have a free domain name request pending review.'))
            raise ValidationError(_('A free domain name has already been approved for this database. It will be online shortly.'))

        desc = self.env['ir.ui.view']._render_template('openerp_enterprise.free_domain_name_support_ticket', {
            'fqdn': fqdn,
            'db': database.db_name,
            'db_uuid': database.db_uuid,
            'email': owner_data['email'],
        })
        task = self.env['project.task'].with_context(mail_notify_author=True).sudo().create({
            'name': f'Free domain name request: {fqdn}',
            'project_id': self.env.ref('openerp_enterprise.project_domain_names').id,
            'partner_id': self.env.user.partner_id.id,
            'tag_ids': [(4, self.env.ref('openerp_enterprise.project_tag_dns').id, None)],
            'user_ids': False,  # Unassigned
            'description': desc,
        })

        req = self.sudo().create({
            'fqdn': fqdn,
            'user_id': self.env.uid,
            'database_id': database.id,
            'state': 'pending',
            'payload': json.dumps(owner_data),
        })

        db_url = f'{SCHEME}://{database.db_name}.{SAAS_BASE_URL}'
        msg_body = Markup(
            '<a href="#" data-oe-model="%s" data-oe-id="%d">Request (ID %d)</a> created.<br>'
            'Website: <a href="%s">%s</a>'
        ) % (req._name, req.id, req.id, db_url, db_url)
        task.message_post(body=msg_body, subtype_xmlid='mail.mt_note')

        return req

    def action_grant(self):
        for req in self:
            payload = json.loads(req.payload)
            contact_type = payload.pop('type')
            fname = payload.pop('given')
            lname = payload.pop('family')
            email = payload.pop('email')
            phone = payload.pop('phone')
            country = payload.pop('country')
            address = payload.pop('streetaddr')
            city = payload.pop('city')

            try:
                req.database_id.allow_free_domain = True
                Domain = self.env['openerp.enterprise.domain'].with_user(req.user_id)
                Domain._register_domain(
                    req.database_id.id,
                    req.fqdn,
                    contact_type,
                    fname,
                    lname,
                    email,
                    phone,
                    country,
                    address,
                    city,
                    **payload
                )
                req.state = 'granted'
            except GandiError as e:
                raise ValidationError("\n -".join([err['description'] for err in e.errors])) from e

    def action_deny(self):
        for req in self:
            req.state = 'denied'


class SpamChecker(models.Model):
    _inherit = 'openerp.enterprise.domain.request'

    def message_update(self, msg, update_vals=None):
        # HOW TO call this code:
        #   NEED TO CREATE AN EMAIL ALIAS for consts.ABUSE_ALIAS
        #   WITH MODEL 'openerp.enterprise.database.request'
        #   WITH a random Record Thread ID set
        if any({dst.split('@')[0] == ABUSE_ALIAS for dst in msg.get('to', '').split(',')}):
            msg['author_id'] = False  # remove author to don't become followers - whatever if domain to check
            sa = self.env.ref("openerp_enterprise.spam_action_server")
            domain_to_checks = sa.sudo().with_context(
                findall=re.findall,
                msg_dict=msg,
            ).run()
            if isinstance(domain_to_checks, str):  # SA return list of url concatened by ,
                headers = {
                    'API-Key': ABUSE_URLSCAN_APIKEY,
                    'Content-Type': 'application/json'
                }
                _logger.info("SpamChecker: Domain to check: %s", domain_to_checks)
                for url in domain_to_checks.split(','):
                    data = {"url": "https://%s" % url, "visibility": "public"}
                    response = requests.post('https://urlscan.io/api/v1/scan/', headers=headers, data=json.dumps(data))
                    _logger.info(response.text)

                # keep minimal info in log
                msg['subject'] = ''
                msg['body'] = domain_to_checks
                msg.pop('attachments', None)


            return True
        else:
            return super().message_update(msg, update_vals=update_vals)


class OpenerpEnterpriseDomain(models.Model):
    _name = 'openerp.enterprise.domain'
    _description = 'Domain names'
    _inherit = ['mail.thread']
    _rec_name = 'fqdn'

    fqdn = fields.Char('FQDN', required=True, readonly=True, index=True, help='Fully Qualified Domain Name')
    expiration_date = fields.Date('Expiration Date', required=True)
    delete_date = fields.Date('Delete Date', required=True, help='End of pending delete period. The domain is released the following day.')
    user_id = fields.Many2one('res.users', 'Owner', index=True, readonly=True, ondelete='set null')
    database_id = fields.Many2one('openerp.enterprise.database', required=True, index=True)
    subscription_id = fields.Many2one(related='database_id.subscription_id', store=True, index=True)
    synced = fields.Boolean('Is synchronized', help='True when data is synchronized with the registrar.')
    state = fields.Selection([
        ('registered', 'Registered'),
        ('integrated', 'Integrated'),
        ('error', 'Error'),
    ], string='State', required=True, default='registered')
    email_confirmed = fields.Boolean('Email Confirmed', help='True when owner email address is confirmed at the registrar level.')
    active = fields.Boolean('Active', default=True)
    renew_order_id = fields.Many2one('sale.order', 'Renewal Order', readonly=True, help='The current renewal order.')
    renew_order_ids = fields.One2many('sale.order', 'domain_id', string='Renew Orders', help='All renewal orders.')
    renew_order_count = fields.Integer("Renew Orders Count", compute='_compute_renew_order_count')
    to_renew = fields.Boolean('To Renew', readonly=True, copy=False)

    def init(self):
        self.env.cr.execute("CREATE UNIQUE INDEX IF NOT EXISTS openerp_enterprise_domain_subscription_id_unique ON %s (subscription_id) WHERE subscription_id IS NOT NULL" % self._table)

    _sql_constraints = [
        ('fqdn_unique', 'unique(fqdn)', 'FQDN must be unique'),
        ('database_id_unique', 'unique(database_id)', 'Database must be unique'),
    ]

    def _compute_renew_order_count(self):
        count = self.env['sale.order'].sudo().read_group(
            [('domain_id', 'in', self.ids)],
            ['domain_id'],
            ['domain_id'],
        )
        count_by_id = {row['domain_id'][0]: row['domain_id_count'] for row in count}
        for domain in self:
            domain.renew_order_count = count_by_id.get(domain.id, 0)

    @api.model
    def _registration_status(self, database):
        # No edu, etc.
        if database.db_name.startswith(SPECIAL_PREFIXES):
            return 'closed'

        # Has website?
        apps = database.app_ids.filtered('installed').mapped('module_id.technical_name')
        website = 'website' in apps

        # No domain registered yet?
        domain = self.with_context(active_test=False).sudo().search([
            '|',
            ('database_id', '=', database.id),
            '&',
            ('user_id', '=', self.env.uid),
            ('subscription_id', '=', False),
        ])

        return 'open' if website and not domain else 'closed'

    @api.model
    def tld(self, fqdn):
        info = PSL.extract(fqdn)
        if info is None:
            return ''

        subdomain, domain, tld = info
        return tld

    @api.model
    def _search_domain(self, search):
        results = {}
        has_tld = '.' in search

        info = PSL.extract(search)
        if info is None:
            domain = search
            etld = ''
        else:
            subdomain, domain, etld = info

        def availability(fqdn):
            try:
                status = self._status(fqdn)
                return status
            except Exception:
                return 'error'

        # If the user chose a TLD, suggest it first.
        if has_tld:
            status = availability(f'{domain}.{etld}')
            results[search] = status

        # Suggest some common TLDs.
        for tld in ['com', 'net', 'org']:
            if etld != tld:
                sld = f'{domain}.{tld}'
                status = availability(sld)
                if status in ['available', 'error']:
                    results[sld] = status

        # Return other possible suggestions among gTLDs.
        excl = set(['.com', '.net', '.org', f'.{etld}'])
        gtlds = [tld for tld in GTLDS if tld not in excl]
        random.shuffle(gtlds)
        for gtld in gtlds:
            sld = f'{domain}{gtld}'
            results[sld] = 'pending'

        return results

    @api.model
    def _verify_custom_domain(self, domain, db):
        if not re.match(r'^[a-z\d][a-z\d-]*\.[a-z\d][a-z\d.-]*\.[a-z]+$', domain):
            return 1

        try:
            cmd = ['dig', '+short', 'CNAME', domain]
            out = subprocess.check_output(cmd, stderr=subprocess.STDOUT, timeout=3, text=True).strip()
        except subprocess.CalledProcessError as e:
            _logger.exception('_verify_custom_domain: failed to run %s: rc=%s, output=%s', cmd, e.returncode, e.output)
            return 2
        except Exception:
            _logger.exception('_verify_custom_domain: unexpected error: cmd=%s', cmd)
            return 2
        else:
            if out == '':  # No CNAME.
                return 3
            if out != f'{db}.{SAAS_BASE_URL}.':  # CNAME pointing elsewhere.
                return 4
            return 0

    @api.model
    def _status(self, fqdn):
        if not fqdn:
            raise ValueError(_('A fully qualified domain name is required.'))

        if len(fqdn.split('.')) < 2:
            raise ValueError(_('A top-level domain is required (e.g. %(fqdn)s.com).', fqdn=fqdn))

        # Check against allowed TLDs.
        if not fqdn.endswith(tuple(TLDS)):
            raise ValidationError(_('TLD not supported.'))

        # Reject IDNs.
        if not all(label.isascii() for label in fqdn.split('.')):
            raise ValueError("IDN not supported.")

        Rules = self.env['openerp.enterprise.trial.rules'].sudo()
        valid_name = Rules._check_dbname(fqdn, scopes=(False, 'dns'))
        if not valid_name:
            raise ValidationError(_('Domain name not allowed.'))

        try:
            response = GANDI.status(fqdn)
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 400:
                raise ValueError('Invalid domain name format.') from e
            raise

        try:
            pricing = response['products'][0]['prices'][0]
            # Ensure premium domains are rejected.
            if pricing.get('type', '') == 'premium':
                raise Exception('Premium domain names are not supported.')
            # Ensure price is within our treshold.
            # normal_price_after_taxes: pricing after tax is applied, when no discount applies (optional).
            # price_after_taxes: pricing after tax is applied.
            price = pricing.get('normal_price_after_taxes', pricing['price_after_taxes'])
            min_duration = pricing.get('min_duration', 1)
            if price * min_duration > PRICE_LIMIT:
                raise Exception('This domain is not covered by our free plan.')
            return response['products'][0]['status']
        except (KeyError, IndexError):
            # Gandi will return 200 OK with e.g. example.com.org but the response body will not contain any domain information.
            return 'unavailable'

    @api.model
    def _register_domain(self, database_id, fqdn, contact_type, fname, lname, email, phone, country, address, city, **kwargs):
        _logger.info('_register_domain: user #%s attempts to register %s with database #%s', self.env.uid, fqdn, database_id)

        fqdn = fqdn.lower()
        fqdn = PSL.sld(fqdn) or fqdn

        # Call status to recheck for supported TLDs premium domains, price cap, block list...
        if self.env['openerp.enterprise.domain']._status(fqdn) != 'available':
            raise ValidationError(_('%(fqdn)s is unavailable.', fqdn=fqdn))

        DB = self.env['openerp.enterprise.database'].sudo()
        databases = [db['db_id'] for db in DB._get_instances(alive=True) if db['is_admin'] and db['is_saas']]
        db = DB.browse(database_id)

        if not db or db.id not in databases:
            raise ValidationError(_('Invalid database.'))

        if self._registration_status(db) != 'open':
            raise ValidationError(_('Unauthorized registration.'))

        owner_data = dict(kwargs, **{
            'type': contact_type,
            'given': fname,
            'family': lname,
            'email': email,
            'phone': phone,
            'country': country,
            'streetaddr': address,
            'city': city,
        })

        sub = db.subscription_id
        Domain = self.with_context(active_test=False).sudo()

        if not db.extra_apps or (sub and sub.subscription_state in ('1_draft', '2_renewal')):
            if Domain.search_count([('database_id', '=', db.id)]):
                raise ValidationError(_('This database is already associated to a domain name.'))
            # One-app free databases are reviewed and allowed on a case-by-case basis.
            if not db.allow_free_domain:
                req = self.env['openerp.enterprise.domain.request']._submit(fqdn, db, owner_data)
                _logger.info('_register_domain: free domain request #%s submitted: fqdn=%s db=%s', req.id, fqdn, db.id)
                return req
        else:
            # An enterprise subscription is required to register a free domain name.
            subscriptions = self.env['sale.order']._get_enterprise_subscriptions()
            if (not db.subscription_id or db.subscription_id not in subscriptions):
                raise ValidationError(_('Your Odoo Subscription is not valid. Go to your subscription or contact us.'))
            # Restrict to one domain name per subscription.
            if Domain.search_count([('subscription_id', '=', db.subscription_id.id)]):
                raise ValidationError(_('Subscription %(code)s is already in use.', code=db.subscription_id.client_order_ref))

        if Domain.search_count([('fqdn', '=', fqdn)]):
            _logger.info('_register_domain: already exists in openerp: domain=%s', fqdn)
            raise ValidationError(_('%(fqdn)s is unavailable.', fqdn=fqdn))

        GANDI.register_domain(fqdn, owner_data, dry_run=TEST)

        # Defaults are set and the record will be synchronized later.
        expiration_date = fields.Date.add(fields.Date.today(), months=11)
        domain = self.sudo().create({
            'fqdn': fqdn,
            'expiration_date': expiration_date,
            # A domain is usually deleted 80 days after the expiration date but it depends on the registry.
            # To avoid making an inaccurate guess we match the expiration date, it will be synchronized later.
            'delete_date': expiration_date,
            'user_id': self.env.uid,
            'database_id': db.id,
            'synced': False,
            'state': 'registered',
        })

        _logger.info('_register_domain: %s successfully registered: id=%s', domain.fqdn, domain.id)

        return domain

    @api.model
    def _set_defaults(self):
        self.ensure_one()
        ok = True

        # Delete all default records set by Gandi.
        try:
            self._delete_dns_records()
        except Exception:
            ok = False
            _logger.exception('_set_defaults: failed to delete dns records: domain=%s', self.fqdn)

        # Redirect the apex domain to the www subdomain.
        try:
            self._create_web_redirection(self.fqdn, f'https://www.{self.fqdn}', 'http301')
        except Exception:
            ok = False
            _logger.exception('_set_defaults: failed to create http redirection: domain=%s', self.fqdn)

        # Create a CNAME record pointing the www subdomain to the database's canonical name.
        try:
            self._put_dns_record('CNAME', 'www', [f'{self.database_id.db_name}.{SAAS_BASE_URL}.'])
        except Exception:
            ok = False
            _logger.exception('_set_defaults: failed to create www CNAME: domain=%s, database=%s', self.fqdn, self.database_id.db_name)

        # Meta: add domain.
        try:
            hostname = f'www.{self.fqdn}'
            self._route_hostname_to_db(hostname)
        except Exception:
            ok = False
            _logger.exception('_set_defaults: _route_hostname_to_db failed: domain=%s, hostname=%s db=%s', self.fqdn, self.database_id.db_name, hostname)

        # Tag the domain with 'saas' so that we can make a distinction between customers domains and our own internal domains.
        try:
            self._tag(['saas'])
        except Exception:
            ok = False
            _logger.exception('_set_defaults: _tag failed: domain=%s', self.fqdn)

        self.state = 'integrated' if ok else 'error'

    @api.model
    def _sync(self):
        """
        The registration is made with minimal data and with inacurate expiration/deletion dates.
        _sync will synchronize the data we have with the registrar's.
        Note that expiration/deletion dates can only be updated to a later date. This is done to
        adapt to the eventually consistent API responses from Gandi (we don't see the new
        expiration date right away), which can be problematic following renew operations.
        """
        self.ensure_one()

        try:
            # Try to get the exact expiration/delete date from the registar.
            info = GANDI.info(self.fqdn)
            expiration_date = fields.Date.to_date(info['dates']['hold_begins_at'])
            self.expiration_date = max(self.expiration_date, expiration_date)
            delete_date = fields.Date.to_date(info['dates']['pending_delete_ends_at'])
            self.delete_date = max(self.delete_date, delete_date)
            # One of: "pending", "done", "failed", "deleted", "none".
            self.email_confirmed = True if info['contacts']['owner']['reachability'] == 'done' else False
            self.synced = expiration_date == self.expiration_date  # retry later
            return True
        except Exception:
            _logger.warning('_sync: data could not be synced: domain=%s', self.fqdn)
            return False

    def _route_hostname_to_db(self, hostname):
        # Inserts hostname into the domains table (meta).
        # This will make the proxy aware that requests to hostname should be routed to db.
        self.ensure_one()
        url = f'{SCHEME}://{self.database_id.db_name}.{SAAS_BASE_URL}/saas_worker/domain-names/trusted-add/'
        payload = {'db_name': self.database_id.db_name, 'hostname': hostname}
        response = requests.post(url, data=payload, timeout=TIMEOUT, auth=SaasAuth())
        response.raise_for_status()
        return True

    def _tag(self, tags):
        self.ensure_one()
        GANDI.tag_domain(self.fqdn, tags)

    def _get_dns_records(self):
        self.ensure_one()

        response = GANDI.get_records(self.fqdn)

        return [{
            'type': rec['rrset_type'],
            'name': rec['rrset_name'],
            'content': content,
            'ttl': rec['rrset_ttl'],
        } for rec in response for content in rec['rrset_values']]

    def _get_record_content(self, rtype, name):
        self.ensure_one()

        if rtype not in ['CNAME', 'MX', 'TXT']:
            raise ValueError(_('Only CNAME, MX and TXT records are supported.'))

        response = GANDI.get_record_content(self.fqdn, rtype, name)

        return response

    def _put_dns_record(self, rtype, name, content, ttl=None):
        self.ensure_one()

        if rtype not in ['CNAME', 'MX', 'TXT']:
            raise ValueError(_('Only CNAME, MX and TXT records are supported.'))

        if ttl and int(ttl) < 300:
            raise ValueError(_('The minimum TTL value is 300 seconds.'))

        GANDI.put_record(self.fqdn, rtype, name, content, ttl=ttl)

    def _delete_dns_records(self):
        self.ensure_one()

        GANDI.delete_records(self.fqdn)

    def _get_contacts(self):
        self.ensure_one()

        response = GANDI.get_contacts(self.fqdn)

        # Transform GANDI's response to expose a subset of attributes only.
        contacts = {contact_type: {
            'fname': contact.given,
            'lname': contact.family,
            'email': contact.email,
            'phone': contact.phone,
        } for contact_type, contact in response.items()}

        return contacts

    def _create_web_redirection(self, host, redirect_to, redirect_type):
        self.ensure_one()
        self._check_redirection(host, redirect_type)
        response = GANDI.create_web_redirection(self.fqdn, host, redirect_to, redirect_type)
        return response

    def _check_redirection(self, host, redirect_type):
        # Provided host should be part of self.fqdn.
        if not host.endswith(self.fqdn):
            raise ValueError(_('The address %(host)s is not part of the domain %(fqdn)s.', host=host, fqdn=self.fqdn))

        # Registrar also allows cloak redirects but it's not really recommended as they mask the real URL.
        if redirect_type not in ['http301', 'http302']:
            raise ValueError(_('Invalid redirection type. Use HTTP 301 or HTTP 302.'))

    def _renew_price(self):
        """
        Get the renew price for 1 year or min_duration (for example .gr
        requires at least 2 years).
        """
        self.ensure_one()
        response = GANDI.status(self.fqdn, process='renew')
        pricing = response['products'][0]['prices'][0]
        price = pricing['price_before_taxes']
        min_duration = pricing.get('min_duration', 1)
        return price * min_duration

    def _quote_renewal(self, auto_commit=True):
        """
        Create a quotation for a 1-year renewal. The price is determined by the
        registrar. Since contacting the registrar's API is time-consuming and
        considering the network is unreliable a transaction is commited every
        loop.
        The quotation is valid until the domain expiration date and requires
        payment in order to be confirmed.
        """
        product = self.env.ref('openerp_enterprise.product_domain_name_renewal')
        for domain in self:
            if domain.renew_order_id:
                # An existing quotation means the renew process is ongoing.
                continue

            price_eur = 0
            try:
                price_eur = domain._renew_price()
            except:
                _logger.exception('failed to get renew price for %s', domain.fqdn)
                continue

            assert price_eur > 0

            company = self.env.ref('base.main_company')  # force renew quotation in Odoo S.A.
            pl = domain.user_id.partner_id.property_product_pricelist
            # Since Odoo S.A. only accepts payments in EUR or USD through Ogone,
            # force USD when the customer currency is not one of these two.
            if pl.currency_id.name not in ('EUR', 'USD'):
                pl = self.env.ref('openerp_enterprise.pl22_america_north')
            price_unit = self.env.ref('base.EUR')._convert(
                price_eur, pl.currency_id, company, fields.Date.today(),
            )

            sale_order = self.env['sale.order'].create({
                'partner_id': domain.user_id.partner_id.id,
                'domain_id': domain.id,
                'order_line': [fields.Command.create({
                    'product_id': product.id,
                    'name': _('Domain name renewal: %s') % (domain.fqdn),
                    'price_unit': price_unit,
                    'product_uom_qty': 1,
                })],
                'pricelist_id': pl.id,
                'company_id': company.id,
                'require_payment' : True,
                'validity_date': domain.expiration_date,
                'payment_term_id': self.env.ref('openerp_enterprise.payment_term_immediate').id,
            })
            domain.renew_order_id = sale_order

            # Notify customer.
            email_ctx = domain.renew_order_id.action_quotation_send().get('context', {})
            template = self.env["mail.template"].browse(email_ctx.get('default_template_id'))
            domain.renew_order_id.with_context(**email_ctx).message_post_with_source(
                template,
                subtype_xmlid='mail.mt_comment',
                email_layout_xmlid='mail.mail_notification_layout_with_responsible_signature',
            )

            if auto_commit:
                self.env.cr.commit()

    def _send_renewal_reminder(self):
        # Don't send reminder for domains with paid quotation and pending renewal.
        domains = self.filtered(lambda d: d.renew_order_id and not d.to_renew)
        for domain in domains:
            domain.renew_order_id.message_post_with_source('openerp_enterprise.reminder_renewal_gandi')

    def _renew(self, auto_commit=True):
        """
        Renew the domain names.

        The renewal deadline for each extension is determined by the registry.
        Example: .com deadlines:
        - Renewal possible until expiration
        - Late renewal possible until 40 days after expiration
        - Deleted by registrar 40 days after expiration
        """
        for domain in self.filtered('to_renew'):
            try:
                if GANDI.renew_domain(domain.fqdn, dry_run=TEST):
                    _logger.info('_renew: domain %s has been renewed', domain.fqdn)

                    domain.expiration_date = fields.Date.add(domain.expiration_date, months=11)
                    domain.delete_date = domain.expiration_date
                    domain.synced = False
                    domain.renew_order_id = False
                    domain.to_renew = False
                    if auto_commit:
                        self.env.cr.commit()
            except Exception:
                _logger.exception('_renew: request to registrar failed: domain=%s', domain.fqdn)

    def action_resend_confirmation_email(self):
        self.ensure_one()
        if not self.env.user.has_group('openerp_enterprise.group_support_functional'):
            raise AccessDenied('Restricted to functional/technical support staff')
        self.sudo()._resend_confirmation_email()

    def _resend_confirmation_email(self):
        for domain in self:
            # Odoo email
            try:
                template_reminder = self.env.ref('openerp_enterprise.reminder_confirm_email_address_gandi')
                email_ctx = dict(self._context, **{
                    'email_to': domain.user_id.email,
                })
                template_reminder.with_context(email_ctx).send_mail(domain.id)
            except Exception:
                _logger.exception('_resend_confirmation_email: failed to send odoo confirmation email: domain=%s', domain.fqdn)

            # Gandi email
            try:
                _logger.info('_resend_confirmation_email: sending gandi email: domain=%s', domain.fqdn)
                GANDI.resend_reachability_email(domain.fqdn)
            except Exception:
                _logger.exception('_resend_confirmation_email: failed to send gandi confirmation email: domain=%s', domain.fqdn)

    def action_quote_renewal(self):
        if not self.env.user.has_group('openerp_enterprise.group_support_functional'):
            raise AccessDenied('Restricted to functional/technical support staff')
        self.sudo()._quote_renewal()

    @api.model
    def _cron_renew_domains(self, auto_commit=True):
        """
        Cron method that:
        - creates renewal quotations for domain names expiring soon
        - sends renewal reminders
        - renews domain names whose quotation has been paid
        """
        # Quote renewals.
        to_quote = self.env['openerp.enterprise.domain'].search([
            ('synced', '=', True),
            ('expiration_date', '<=', fields.Date.add(fields.Date.today(), days=45)),
            ('renew_order_id', '=', False),
            ('to_renew', '=', False),
            ('user_id', '!=', False),
        ])
        to_quote._quote_renewal(auto_commit=auto_commit)

        # Send renewal reminders at D-30, D-14 and D-7.
        for days in [30, 14, 7]:
            reminder_date = fields.Date.add(fields.Date.today(), days=days)
            to_remind = self.env['openerp.enterprise.domain'].search([
                ('expiration_date', '>=', reminder_date),
                ('expiration_date', '<', fields.Date.add(reminder_date, days=1)),
                ('renew_order_id', '!=', False),
                ('to_renew', '=', False),
            ])
            to_remind._send_renewal_reminder()

        # Renew domains.
        to_renew = self.env['openerp.enterprise.domain'].search([
            ('synced', '=', True),
            ('renew_order_id.state', '=', 'sale'),
            ('to_renew', '=', True),
        ])
        to_renew._renew(auto_commit=auto_commit)

    @api.model
    def _cron_sync_data_with_registrar(self, auto_commit=True):
        domains = self.env['openerp.enterprise.domain'].search([
            '|', ('synced', '=', False),
            ('email_confirmed', '=', False),
        ])

        for domain in domains:
            domain._sync()
            if auto_commit:
                self.env.cr.commit()

    @api.model
    def _cron_set_default_dns_config(self, auto_commit=True):
        domains = self.env['openerp.enterprise.domain'].search([
            ('state', '=', 'registered'),
        ])

        for domain in domains:
            domain._set_defaults()
            if auto_commit:
                self.env.cr.commit()

    @api.model
    def _cron_resend_confirmation_email(self):
        # Resend the confirmation email after 5, 10 and 14 days.
        for days in [5, 10, 14]:
            # Fetch domains created days ago.
            domains = self.env['openerp.enterprise.domain'].search([
                ('email_confirmed', '=', False),
                ('synced', '=', True),
                ('create_date', '>=', fields.Date.subtract(fields.Date.today(), days=days)),
                ('create_date', '<', fields.Date.subtract(fields.Date.today(), days=days - 1)),
            ])

            domains._resend_confirmation_email()

    @api.model
    def _get_domain(self, fqdn):
        return self.sudo().search([
            ('user_id', '=', self.env.uid),
            ('fqdn', '=', fqdn),
        ], limit=1)

    @api.model
    def _get_user_domains(self, database_id):
        """ Return current user domains."""
        return self.sudo().search([
            ('user_id', '=', self.env.uid),
            ('database_id', '=', database_id),
        ])

    @api.autovacuum
    def _gc_domains(self):
        domains = self.search([
            ('synced', '=', True),
            ('delete_date', '<', fields.Date.today()),
        ])
        _logger.info('_gc_domains: %s.', ', '.join(domains.mapped('fqdn')))
        domains.active = False

    def action_view_renewal_orders(self):
        self.ensure_one()
        return {
            'name': _('Renewal Orders'),
            'type': 'ir.actions.act_window',
            'view_mode': 'tree,form',
            'res_model': 'sale.order',
            'domain': [('id', 'in', self.renew_order_ids.ids)],
        }


class SaleSubscription(models.Model):
    _inherit = 'sale.order'

    domain_id = fields.Many2one('openerp.enterprise.domain', 'Domain name')

    @api.model
    def _get_enterprise_subscriptions(self):
        tag_ids = [
            self.env.ref('openerp_enterprise.standard_plan_tag').id,
            self.env.ref('openerp_enterprise.custom_plan_tag').id,
        ]
        partner = self.env.user.partner_id.commercial_partner_id
        subs = self.env['sale.order'].sudo().search([
            ('is_subscription', '=', True),
            '|',
            ('enterprise_final_customer_id', 'child_of', partner.id),
            ('partner_id', 'child_of', partner.id),
            ('sale_order_template_id.subscription_tag_ids', 'in', tag_ids),
            ('subscription_state', 'in', ['3_progress', '4_paused']),
        ])

        return subs


class AccountMove(models.Model):
    _inherit = 'account.move'

    def _invoice_paid_hook(self):
        """
        Sale order confirmation is not enough. Full payment is required for
        domain renewals.
        """
        res = super(AccountMove, self)._invoice_paid_hook()
        invoices = self.filtered(lambda move: move.move_type == 'out_invoice' and move.state == 'posted')
        for invoice in invoices:
            domains = invoice.invoice_line_ids.mapped('sale_line_ids.order_id.domain_id')
            domains.sudo().write({'to_renew': True})
        return res
